# Copyright 2020, feelSpace GmbH, <info@feelspace.de>
import logging
from typing import List, Optional


class GattAttribute:

    def __init__(self, handle):
        """
        Gatt attribute constructor.
        :param Optional(int) handle: Attribute handle.
        """
        self.handle = handle

    def __eq__(self, other):
        return self.handle == other.handle

    def __ne__(self, other):
        return not self.__eq__(other)


class GattCharacteristic:

    def __init__(self, uuid, declaration_attr, value_attr, configuration_attrs=None):
        """
        Gatt characteristic constructor.
        :param str uuid: Characteristic UUID.
        :param GattAttribute declaration_attr: Declaration attribute.
        :param GattAttribute value_attr: Value attribute
        :param List[GattAttribute] configuration_attrs: Configuration attributes.
        """
        if configuration_attrs is None:
            configuration_attrs = []
        self.uuid = uuid
        self.declaration_attr = declaration_attr
        self.value_attr = value_attr
        self.configuration_attrs: List[GattAttribute] = configuration_attrs

    def __eq__(self, other):
        return self.uuid == other.uuid

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_all_handles(self) -> [int]:
        """ Returns the list of all attribute handles covered by this characteristic.
        :return: the list of all attribute handles covered by this characteristic.
        """
        handles = []
        if self.declaration_attr.handle is not None:
            handles.append(self.declaration_attr.handle)
        if self.value_attr.handle is not None:
            handles.append(self.value_attr.handle)
        if self.configuration_attrs is not None:
            for attr in self.configuration_attrs:
                if attr.handle is not None:
                    handles.append(attr.handle)
        return handles


class NaviBeltGattProfile:

    def __init__(self):
        """ Constructor of a NaviBelt GATT profile.
        """

        self.logger = logging.getLogger(__name__)

        self.firmware_info_char = GattCharacteristic(
            "0000fe01-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None))
        self.keep_alive_char = GattCharacteristic(
            "0000fe02-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None), [GattAttribute(None)])
        self.vibration_command_char = GattCharacteristic(
            "0000fe03-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None))
        self.button_press_char = GattCharacteristic(
            "0000fe04-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None), [GattAttribute(None)])
        self.param_request_char = GattCharacteristic(
            "0000fe05-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None))
        self.param_notification_char = GattCharacteristic(
            "0000fe06-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None), [GattAttribute(None)])
        self.battery_status_char = GattCharacteristic(
            "0000fe09-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None), [GattAttribute(None)])
        self.sensor_command_char = GattCharacteristic(
            "0000fe0a-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None))
        self.sensor_notification_char = GattCharacteristic(
            "0000fe0b-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None), [GattAttribute(None)])
        self.orientation_data_char = GattCharacteristic(
            "0000fe0c-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None), [GattAttribute(None)])
        self.debug_input_char = GattCharacteristic(
            "0000fe13-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None))
        self.debug_output_char = GattCharacteristic(
            "0000fe14-0000-1000-8000-00805f9b34fb", GattAttribute(None), GattAttribute(None), [GattAttribute(None)])

        self.characteristics = [
            self.firmware_info_char,
            self.keep_alive_char,
            self.vibration_command_char,
            self.button_press_char,
            self.param_request_char,
            self.param_notification_char,
            self.battery_status_char,
            self.sensor_command_char,
            self.sensor_notification_char,
            self.orientation_data_char,
            self.debug_input_char,
            self.debug_output_char
        ]

        self._handles_dict = {}
        for char in self.characteristics:
            for handle in char.get_all_handles():
                self._handles_dict[handle] = char

        self._char_uuid_dict = {}
        for char in self.characteristics:
            self._char_uuid_dict[char.uuid.lower()] = char

    def set_char_handles(self, char_uuid, declaration_handle, value_handle, configuration_handles=None):
        """ Sets the attributes handles for a characteristic.

        :param char_uuid: The characteristic uuid.
        :param declaration_handle: The declaration attribute handle.
        :param value_handle: The value attribute handle.
        :param configuration_handles: The configuration attribute handles.
        """
        if configuration_handles is None:
            configuration_handles = []
        characteristic = self.get_char_from_uuid(char_uuid)
        if characteristic is None:
            self.logger.error("NaviBeltGattProfile: No characteristic to set handles.")
            return
        if len(characteristic.configuration_attrs) != len(configuration_handles):
            self.logger.error("NaviBeltGattProfile: Non-matching length for configuration handles.")
            return
        # Remove previous handles from dict
        for handle in characteristic.get_all_handles():
            self._handles_dict.pop(handle, None)
        # Set new handles
        characteristic.declaration_attr.handle = declaration_handle
        characteristic.value_attr.handle = value_handle
        for idx, conf_attr in enumerate(characteristic.configuration_attrs):
            conf_attr.handle = configuration_handles[idx]
        # Adds handles to dict
        for handle in characteristic.get_all_handles():
            self._handles_dict[handle] = characteristic

    def get_char_from_handle(self, handle) -> Optional[GattCharacteristic]:
        """ Returns the GATT characteristic that contains the given attribute handle.

        :param int handle: The attribute handle.
        :return: The characteristic that contains the attribute.
        """
        if handle in self._handles_dict:
            return self._handles_dict[handle]
        return None

    def get_char_from_uuid(self, uuid) -> Optional[GattCharacteristic]:
        """ Returns the characteristic from its UUID.

        :param str uuid: The characteristic UUID.
        :return: The characteristic.
        """
        if uuid.lower() in self._char_uuid_dict:
            return self._char_uuid_dict[uuid]
        return None


_usb_gatt_profile = None


def get_usb_gatt_profile() -> NaviBeltGattProfile:
    """
    Returns the GATT profile for a USB connection (no service discovery).
    :return: The GATT profile for a USB connection.
    """
    global _usb_gatt_profile
    if _usb_gatt_profile is None:
        _usb_gatt_profile = NaviBeltGattProfile()
        # Handles initialization
        _usb_gatt_profile.set_char_handles("0000fe01-0000-1000-8000-00805f9b34fb", 30, 31)
        _usb_gatt_profile.set_char_handles("0000fe02-0000-1000-8000-00805f9b34fb", 32, 33, [34])
        _usb_gatt_profile.set_char_handles("0000fe03-0000-1000-8000-00805f9b34fb", 35, 36)
        _usb_gatt_profile.set_char_handles("0000fe04-0000-1000-8000-00805f9b34fb", 37, 38, [39])
        _usb_gatt_profile.set_char_handles("0000fe05-0000-1000-8000-00805f9b34fb", 40, 41)
        _usb_gatt_profile.set_char_handles("0000fe06-0000-1000-8000-00805f9b34fb", 42, 43, [44])
        _usb_gatt_profile.set_char_handles("0000fe09-0000-1000-8000-00805f9b34fb", 50, 51, [52])
        _usb_gatt_profile.set_char_handles("0000fe0a-0000-1000-8000-00805f9b34fb", 54, 55)
        _usb_gatt_profile.set_char_handles("0000fe0b-0000-1000-8000-00805f9b34fb", 56, 57, [58])
        _usb_gatt_profile.set_char_handles("0000fe0c-0000-1000-8000-00805f9b34fb", 59, 60, [61])
        _usb_gatt_profile.set_char_handles("0000fe13-0000-1000-8000-00805f9b34fb", 74, 75)
        _usb_gatt_profile.set_char_handles("0000fe14-0000-1000-8000-00805f9b34fb", 76, 77, [78])
    return _usb_gatt_profile



# Copyright 2020, feelSpace GmbH, <info@feelspace.de>
import asyncio
import queue
import sys
import threading
import time
import logging

import bleak
import serial
from bleak import BleakScanner, BleakClient, BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from pybelt import belt_scanner

from pybelt._gatt_profile import *

SERIAL_BAUDRATE = 115200
# Baudrate for serial connection

SERIAL_READ_TIMEOUT = 0.50
# Timeout to read data on serial port

THREAD_JOIN_TIMEOUT_SEC = 2.0
# Timeout to wait thread termination

SERIAL_PACKET_TIMEOUT = 2.5
# Timeout to read a complete packet on serial port

SERIAL_FLUSH_INPUT_TIMEOUT = 1.50
# Timeout for flushing serial input

EVENT_LOOP_READY_TIMEOUT = 1.0


# Timeout for initializing the event loop in BLE interface


def decode_ascii(byte_msg) -> str:
    """Decodes a byte array to a string.

    :param Union[bytes, bytearray] byte_msg:
        The bytes to decode.

    :rtype str
        The decoded string.
    """
    if len(byte_msg) == 0:
        return ""
    try:
        for i in range(0, len(byte_msg)):
            if byte_msg[i] < 0 or byte_msg[i] > 255:
                byte_msg[i] = 35
        msg = bytes(byte_msg).decode('ascii', 'replace')
        return msg
    except:
        return "### FORMAT ERROR ###"


class ParseMode:
    """Enumeration of parse mode for connection interfaces."""

    GATT = 0
    TEXT = 1


class BeltCommunicationDelegate:
    """
    Delegate interface for communication interfaces.
    """

    def on_connection_established(self):
        """
        Called when the connection has been established.
        """
        pass

    def on_connection_closed(self, expected=True):
        """
        Called when the connection has been closed.

        :param bool expected: 'True' if the disconnection was expected.
        """
        pass

    def on_gatt_char_notified(self, gatt_char, data):
        """
        Called when a GATT notification has been received or a characteristic has been read.

        :param GattCharacteristic gatt_char: The characteristic that received the data.
        :param bytes data: The data received.
        """
        pass

    def on_raw_text_received(self, data):
        """
        Called when raw text has been received.
        This applies only to the serial interface.

        :param bytes data: The data received.
        """
        pass


class BeltCommunicationInterface:
    """
    Interface for the communication.
    """

    def close(self):
        """
        Closes the connection.
        """
        pass

    def wait_disconnection(self, timeout=1) -> bool:
        """
        Waits until the belt disconnect.

        :param float timeout: The timeout time to stop waiting for disconnection.
        :return: 'True' if the belt disconnected, 'False' if the timeout is reached.
        """
        pass

    def is_connected(self) -> bool:
        """
        Returns the connection state.

        :return: 'True' if the belt is connected.
        """
        pass

    def write_gatt_char(self, gatt_char, data) -> bool:
        """
        Writes a GATT attribute.

        :param GattCharacteristic gatt_char:  The characteristic to write.
        :param bytes data: The data to write.
        :return: 'True' if successful, 'False' if not connected or a problem occurs.
        """
        pass

    def set_gatt_notifications(self, gatt_char, enabled) -> bool:
        """
        Enables or disables the notifications for a characteristic.

        :param GattCharacteristic gatt_char: The characteristic for which notifications has to be enabled or disabled.
        :param bool enabled: 'True' to enable the notifications on the characteristic.
        :return: 'True' if successful, 'False' if not connected or a problem occurs.
        """
        pass

    def read_gatt_char(self, gatt_char) -> bool:
        """
        Request the value of a characteristic.

        :param GattCharacteristic gatt_char: The GATT characteristic to read.
        :return: 'True' if successful, 'False' if not connected or a problem occurs.
        """
        pass

    def get_gatt_profile(self) -> NaviBeltGattProfile:
        """
        Returns the GATT profile with characteristics.
        :return: the GATT profile with characteristics.
        """
        pass


class SerialPortInterface(threading.Thread, BeltCommunicationInterface):
    """Serial port interface.
    """

    # --------------------------------------------------------------- #
    # Public methods

    def __init__(self, delegate):
        """Initializes the serial port interface.

        Parameters
        ----------
        :param BeltCommunicationDelegate delegate:
            The delegate that handles packets and messages received.
        """
        threading.Thread.__init__(self, name="SerialPortListener")
        self._serial_port_name = None
        self._serial_port = None
        self._delegate = delegate

        # Logger
        self.logger = logging.getLogger(__name__)

        # Flag for stopping the thread
        self.stop_flag = False
        self._expect_disconnection = False

        # Parse mode
        self._parse_mode = ParseMode.GATT
        self._pending_parse_mode = None

        # Variable for packet timeout
        self._packet_start_time = time.perf_counter()

        # Output lock
        self._output_lock = threading.RLock()

        # Python version
        self._PY3 = sys.version_info > (3,)

        # GATT profile
        self._gatt_profile = get_usb_gatt_profile()

    def open(self, port, parse_mode=ParseMode.GATT, initial_flush=True):
        """
        Opens a serial connection with a belt.
        :param str port: The port name to use.
        :param int parse_mode: The parse mode.
        :param bool initial_flush: 'True' to flush initial input data when connected.
        """
        # Check connection state
        if self._serial_port is not None:
            self.logger.error("SerialPortInterface: Belt already connected!")
            return

        # Set parameters
        self._parse_mode = parse_mode
        self._pending_parse_mode = None

        # Open port
        self._serial_port_name = port
        self.stop_flag = False
        self._serial_port = serial.Serial(
            self._serial_port_name,
            SERIAL_BAUDRATE,
            timeout=SERIAL_READ_TIMEOUT)

        # Initial flush
        self.write_gatt_char(self._gatt_profile.param_request_char, b'\x01\x01')
        if initial_flush:
            self._flush_input()

        # Start listening
        self.start()

        # Dummy command to start binary communication
        self.write_gatt_char(self._gatt_profile.param_request_char, b'\x01\x01')
        time.sleep(0.5)

        # Inform delegate
        if self._delegate is not None:
            try:
                self._delegate.on_connection_established()
            except:
                pass

    def set_parse_mode(self, parse_mode):
        """Sets the parse mode.

        Parameters
        ----------
        :param int parse_mode:
            The parse mode to set.
        """
        if self.is_alive() and not self.stop_flag and self._parse_mode != parse_mode:
            self._pending_parse_mode = parse_mode

    def get_port(self) -> str:
        """
        Returns the port for the connection.
        :return: the port for the connection.
        """
        return self._serial_port_name

    # --------------------------------------------------------------- #
    # Private methods

    def _flush_input(self):
        """Flushes the input. """
        self.logger.info("SerialPortListener: Flush input.")
        try:
            flush_timeout = time.perf_counter() + SERIAL_FLUSH_INPUT_TIMEOUT
            while time.perf_counter() < flush_timeout:
                self._serial_port.reset_input_buffer()
                time.sleep(0.05)
        except:
            self.logger.warning("SerialPortListener: Unable to flush serial input.")

    # --------------------------------------------------------------- #
    # Implementation of communication interface methods

    def close(self):
        # Stop listener
        self.stop_flag = True
        if self.is_alive() and threading.current_thread() != self:
            self.join(THREAD_JOIN_TIMEOUT_SEC)

    def wait_disconnection(self, timeout=1) -> bool:
        if not self.is_connected():
            # Already disconnected
            return True
        if threading.current_thread() == self:
            self.logger.error("BeltController: Cannot wait disconnection from listener thread.")
            return False
        self._expect_disconnection = True
        self.join(timeout)
        self._expect_disconnection = False
        return self.is_connected()

    def is_connected(self) -> bool:
        return self._serial_port is not None

    def write_gatt_char(self, gatt_char, data) -> bool:
        with self._output_lock:
            try:
                packet = bytes([gatt_char.value_attr.handle]) + bytes([len(data)]) + data
                self._serial_port.write(packet)
            except:
                return False
        return True

    def set_gatt_notifications(self, gatt_char, enabled) -> bool:
        with self._output_lock:
            try:
                if enabled:
                    packet = bytes([gatt_char.configuration_attrs[0].handle, 2, 0x01, 0x00])
                else:
                    packet = bytes([gatt_char.configuration_attrs[0].handle, 2, 0x00, 0x00])
                self._serial_port.write(packet)
            except:
                return False
        return True

    def read_gatt_char(self, gatt_char) -> bool:
        with self._output_lock:
            try:
                packet = bytes([gatt_char.value_attr.handle, 0])
                self._serial_port.write(packet)
            except:
                return False
        return True

    def get_gatt_profile(self) -> NaviBeltGattProfile:
        return self._gatt_profile

    # --------------------------------------------------------------- #
    # Implementation of Thread methods

    def run(self):
        self.stop_flag = False
        self._expect_disconnection = False
        self.logger.debug("SerialPortListener: Start listening belt.")
        packet = None
        while not self.stop_flag:
            try:
                # Blocking until data are received or read timeout
                data_serial = self._serial_port.read(size=1)
                # Convert to list of int
                if len(data_serial) > 0:
                    in_byte = bytes([data_serial[0]])
                else:
                    in_byte = None
            except:
                if not self.stop_flag and not self._expect_disconnection:
                    self.logger.exception("SerialPortListener: Error when reading on serial port.")
                break
            # Check for packet timeout
            if packet is not None and len(packet) > 0:
                if ((time.perf_counter() - self._packet_start_time) >
                        SERIAL_PACKET_TIMEOUT):
                    # Timeout, clear packet
                    self.logger.error("SerialPortListener: Packet timeout. (" + decode_ascii(packet) + ")")
                    packet = None
            # Check parse mode
            if self._pending_parse_mode is not None:
                # Change parse mode
                self._parse_mode = self._pending_parse_mode
                self._pending_parse_mode = None
                # Clear packet
                packet = None
            # Handle received byte (fill packet)
            if in_byte is not None:
                if self._parse_mode == ParseMode.GATT:
                    # Fill packet
                    if packet is None or len(packet) == 0:
                        packet = bytearray()
                        # First byte is attribute handle
                        gatt_char = self._gatt_profile.get_char_from_handle(ord(in_byte))
                        if gatt_char is not None:
                            packet.append(in_byte[0])
                            self._packet_start_time = time.perf_counter()
                        else:
                            self.logger.error("SerialPortListener: Incorrect attribute handle in packet header. (" +
                                              decode_ascii(in_byte) + ")")
                            self._flush_input()
                            # ignore byte
                    elif len(packet) == 1:
                        # Second byte is the data length
                        if ord(in_byte) <= 22:
                            packet.append(in_byte[0])
                        else:
                            gatt_char = self._gatt_profile.get_char_from_handle(packet[0])
                            if gatt_char is not None and gatt_char == self._gatt_profile.sensor_notification_char and \
                                    ord(in_byte) <= 244:
                                # Data length extension supported on sensor notifications
                                packet.append(in_byte[0])
                            else:
                                # Incorrect length, clear received data
                                self.logger.error("SerialPortListener: Incorrect packet length in packet header. (" +
                                                  decode_ascii(in_byte) + ")")
                                self._flush_input()
                                packet = None
                    elif len(packet) >= 2:
                        # Complete packet to the data size
                        packet.append(in_byte[0])
                    # Check for complete packet
                    if packet is not None and (len(packet) >= 2) and (len(packet) == packet[1] + 2):
                        # Notify packet received
                        try:
                            gatt_char = self._gatt_profile.get_char_from_handle(packet[0])
                            self._delegate.on_gatt_char_notified(
                                gatt_char,
                                packet[2:])
                        except:
                            self.logger.exception("SerialPortListener: Error when handling received packet.")
                        packet = None
                elif self._parse_mode == ParseMode.TEXT:
                    if (ord(in_byte) == 0x0D or ord(in_byte) == 0x0A or
                            (32 <= ord(in_byte) <= 127)):
                        if packet is None:
                            packet = bytearray()
                        packet.append(in_byte[0])
                    else:
                        # Ignore byte and clear packet
                        self.logger.error("SerialPortListener: Incorrect text byte. (" +
                                          decode_ascii(in_byte) + ")")
                        packet = None
                    # Check for completed packet (new line)
                    if packet is not None and len(packet) > 0 and packet[-1] == 0x0A:
                        # Handle packet
                        try:
                            self._delegate.on_raw_text_received(packet)
                        except:
                            self.logger.exception("Error when handling received text.")
                        packet = None
        self.logger.debug("SerialPortListener: Stop listening belt.")
        # Close port
        if self._serial_port is not None:
            try:
                self._serial_port.close()
            except:
                self.logger.exception("BeltController: Failed to close serial port.")
        self._serial_port = None
        # Inform delegate
        if self._delegate is not None:
            try:
                self._delegate.on_connection_closed(expected=(self.stop_flag or self._expect_disconnection))
            except:
                pass


def print_packet(packet):
    packet_str = ""
    for b in packet:
        packet_str += hex(b) + " "
    print(packet_str)


class BleInterface(BeltCommunicationInterface, threading.Thread):
    """Serial port interface.
    """

    # --------------------------------------------------------------- #
    # Public methods

    def __init__(self, delegate):
        """Initializes the BLE interface.

        :param BeltCommunicationDelegate delegate:
            The delegate that handles received notifications.
        """
        threading.Thread.__init__(self, name="BleInterfaceThread")
        self._device = None
        self._delegate = delegate
        self._gatt_client = None  # type: Optional[BleakClient]
        self._event_loop = None
        self._event_loop_ready = threading.Event()
        self._event_notifier = None
        self._is_disconnecting = False
        self._expect_disconnection = False
        # Logger
        self.logger = logging.getLogger(__name__)
        # GATT profile
        self._gatt_profile = get_usb_gatt_profile()

    def open(self, device=None):
        """
        Connects to a belt via BLE.

        :param bleak.backends.device.BLEDevice device: The device to connect to, or 'None' to scan for a device.
        """
        if self._gatt_client is not None:
            # Already connected
            self.logger.error("BleInterface: Belt already connected!")
            return
        # Start event notifier
        self._event_notifier = BleEventNotifier(self._delegate, self)
        self._event_notifier.start()
        # Start thread and loop
        self._event_loop_ready.clear()
        self.start()
        self._event_loop_ready.wait(EVENT_LOOP_READY_TIMEOUT)
        # Retrieve device
        self._device = device
        if self._device is None:
            try:
                # Use belt scanner to find the first belt
                belts = []
                with belt_scanner.create() as scanner:
                    belts = scanner.scan()
                if not belts:
                    self._device = belts[0]
            except:
                self.logger.exception("BleInterface: Error when scanning!")
                self.close()
                raise
        if self._device is None:
            raise Exception("No belt found via BLE.")
        try:
            # Connect to device
            future = asyncio.run_coroutine_threadsafe(self._connect(), self._event_loop)
            connected = future.result()
            if not connected or self._gatt_client is None:
                self.close()
                raise Exception("BLE connection failed.")
            # Retrieve profile / Service discovery (automatic)
            self._fill_gatt_profile(self._gatt_client.services)
        except:
            # Disconnect and re-raise exception
            self.logger.exception("BleInterface: Error when scheduling connection!")
            self.close()
            raise
        # Inform delegate
        try:
            self._event_notifier.notify_connection()
        except:
            self.logger.exception("BleInterface: Failed to access event notifier for connection notification!")
            pass

    def get_device(self):
        """
        Returns the connected device.

        :return bleak.backends.device.BLEDevice: The connected device.
        """
        return self._device

    # --------------------------------------------------------------- #
    # Private methods

    async def _connect(self) -> bool:
        """
        Connects the belt.
        :return: 'True' if successful, 'False' otherwise.
        """
        try:
            self.logger.debug("BleInterface: Connect client.")
            self._gatt_client = BleakClient(
                self._device,
                disconnected_callback=self._on_device_disconnected,
                loop=self._event_loop)
            await self._gatt_client.connect()
        except:
            self.logger.exception("BleInterface: Error when connecting!")
            return False
        self.logger.debug("BleInterface: Client connected.")
        return True

    async def _disconnect(self) -> bool:
        """
        Disconnects the device.
        :return: 'True' if successful, 'False' otherwise.
        """
        success = True
        try:
            if self._gatt_client is not None:
                if self._gatt_client.is_connected:
                    self.logger.debug("BleInterface: Disconnect client.")
                    success = await self._gatt_client.disconnect()
                else:
                    self.logger.debug("BleInterface: Client already disconnected.")
        except:
            self.logger.exception("BleInterface: Error when disconnecting!")
            success = False
        return success

    async def _write_gatt_char(self, gatt_char, data) -> bool:
        """
        Writes a GATT characteristic.
        :param GattCharacteristic gatt_char: The characteristic to write.
        :param bytes data: The data to write.
        :return: 'True' if successful, 'False' otherwise.
        """
        try:
            if self._gatt_client is None:
                self.logger.warning("BleInterface: No connection to write char!")
                return False
            if not self._gatt_client.is_connected:
                self.logger.warning("BleInterface: No connection to set notifications!")
                return False
            await self._gatt_client.write_gatt_char(gatt_char.uuid, bytearray(data), response=True)
        except:
            self.logger.exception("BleInterface: Error when writing characteristic.")
            return False
        return True

    async def _set_gatt_notifications(self, gatt_char, enabled) -> bool:
        """
        Enables or disables the notifications.
        :param GattCharacteristic gatt_char: The GATT characteristic to configure.
        :param enabled: 'True' to enable notifications, 'False' to disable notifications.
        :return: 'True' if successful, 'False' otherwise.
        """
        try:
            if self._gatt_client is None:
                self.logger.warning("BleInterface: No connection to set notifications!")
                return False
            if not self._gatt_client.is_connected:
                self.logger.warning("BleInterface: No connection to set notifications!")
                return False
            if enabled:
                await self._gatt_client.start_notify(gatt_char.uuid, self._on_notification_received)
            else:
                await self._gatt_client.stop_notify(gatt_char.uuid)
        except:
            self.logger.exception("BleInterface: Error when configuring notifications!")
            return False
        return True

    async def _read_gatt_char(self, gatt_char) -> bool:
        """
        Reads a GATT characteristic.
        :param GattCharacteristic gatt_char: The GATT characteristic to read.
        :return: 'True' if successful, 'False' otherwise.
        """
        try:
            if self._gatt_client is None:
                self.logger.warning("BleInterface: No connection to read char!")
                return False
            if not self._gatt_client.is_connected:
                self.logger.warning("BleInterface: No connection to set notifications!")
                return False
            value = await self._gatt_client.read_gatt_char(gatt_char.uuid)
        except:
            self.logger.exception("BleInterface: Error when reading characteristic.")
            return False
        try:
            self._event_notifier.notify_gatt_notification(gatt_char, bytes(value))
        except:
            self.logger.exception("BleInterface: Error when calling delegate method 'on_gatt_char_notified'.")
        return True

    def _fill_gatt_profile(self, services):
        """ Fills the gatt profile with attribute handles.
        :param BleakGATTServiceCollection services: The list of services.
        """
        self.logger.debug("BleInterface: Service discovery.")
        for gatt_char in self._gatt_profile.characteristics:
            bleak_gatt_char: BleakGATTCharacteristic = services.get_characteristic(gatt_char.uuid)
            if bleak_gatt_char is None:
                self.logger.error("BleInterface: Characteristic not listed for UUID {}.".format(gatt_char.uuid))
            else:
                descriptor_handles = []
                for descriptor in bleak_gatt_char.descriptors:
                    descriptor_handles.append(descriptor.handle)
                self._gatt_profile.set_char_handles(gatt_char.uuid, bleak_gatt_char.handle, bleak_gatt_char.handle + 1,
                                                    descriptor_handles)

    # --------------------------------------------------------------- #
    # Bleak GATT client callback methods

    def _on_notification_received(self, characteristic, data):
        """
        Callback for notifications.
        :param BleakGATTCharacteristic characteristic: The characteristic.
        :param bytearray data: The characteristic notified value.
        """
        gatt_char = self._gatt_profile.get_char_from_handle(characteristic.handle)
        if gatt_char is None:
            self.logger.debug("BleInterface: Notification on unsupported handle ({})!".format(characteristic.handle))
            return
        try:
            self._event_notifier.notify_gatt_notification(gatt_char, bytes(data))
        except:
            self.logger.exception("BleInterface: Failed to access event notifier!")

    def _on_device_disconnected(self, gatt_client):
        """
        Callback on disconnection.
        :param BleakClient gatt_client: The GATT client disconnected.
        """
        self.logger.debug("BleInterface: Client {} disconnected.".format(gatt_client.address))
        if self._gatt_client is None:
            # Connection already closed
            return
        if self._is_disconnecting:
            # Ignore callback when properly disconnecting from 'close()' call
            return

        # TODO TBT
        # try:
        #     # Disconnect
        #     self._gatt_client.disconnect()
        #     # TODO
        #     #future = asyncio.run_coroutine_threadsafe(self._disconnect(), self._event_loop)
        #     #success = future.result()
        #     if not success:
        #         self.logger.error("BleInterface: Failed to disconnect!")
        # except:
        #     self.logger.exception("BleInterface: Error when scheduling disconnection!")

        # Disconnected
        self._gatt_client = None
        # Stop event loop
        try:
            self.logger.debug("BleInterface: Stop event loop.")
            self._event_loop.call_soon_threadsafe(self._event_loop.stop)
        except:
            self.logger.exception("BleInterface: Error when stopping event loop!")
        self._event_loop = None
        # Notify disconnection and stop event notifier
        if self._event_notifier is not None:
            self._event_notifier.notify_disconnection(expected=self._expect_disconnection)

    # --------------------------------------------------------------- #
    # Implementation of communication interface methods

    def close(self):
        if self._gatt_client is None:
            # Already closed
            return
        self.logger.debug("BleInterface: Disconnect and stop event loop.")
        # Flag to ignore '_on_device_disconnected' callback
        self._is_disconnecting = True
        try:
            # Disconnect
            future = asyncio.run_coroutine_threadsafe(self._disconnect(), self._event_loop)
            success = future.result()
            if not success:
                self.logger.error("BleInterface: Failed to disconnect!")
        except:
            self.logger.exception("BleInterface: Error when scheduling disconnection!")
        self._is_disconnecting = False
        self._gatt_client = None
        # Stop event loop
        try:
            self.logger.debug("BleInterface: Stop event loop.")
            self._event_loop.call_soon_threadsafe(self._event_loop.stop)
        except:
            self.logger.exception("BleInterface: Error when stopping event loop!")
        self._event_loop = None
        # Notify disconnection and stop event notifier
        if self._event_notifier is not None:
            self._event_notifier.notify_disconnection()

    def wait_disconnection(self, timeout=1) -> bool:
        if not self.is_connected():
            # Already disconnected
            return True
        if threading.current_thread() == self:
            self.logger.error("BeltController: Cannot wait disconnection from listener thread.")
            return False
        self._expect_disconnection = True
        self._event_notifier.join(timeout)
        self._expect_disconnection = False
        return self.is_connected()

    def is_connected(self) -> bool:
        return self._gatt_client is not None

    def write_gatt_char(self, gatt_char, data) -> bool:
        if self._gatt_client is None:
            self.logger.error("BleInterface: No connection to write char!")
            return False
        try:
            future = asyncio.run_coroutine_threadsafe(self._write_gatt_char(gatt_char, data), self._event_loop)
            success = future.result()
            if not success:
                self.logger.error("BleInterface: Failed to write char!")
                return False
        except:
            self.logger.exception("BleInterface: Error when scheduling write char operation!")
            return False
        return True

    def set_gatt_notifications(self, gatt_char, enabled) -> bool:
        if self._gatt_client is None:
            self.logger.error("BleInterface: No connection to set notifications!")
            return False
        try:
            future = asyncio.run_coroutine_threadsafe(
                self._set_gatt_notifications(gatt_char, enabled), self._event_loop)
            success = future.result()
            if not success:
                self.logger.error("BleInterface: Failed to set notification!")
                return False
        except:
            self.logger.exception("BleInterface: Error when scheduling notification configuration!")
            return False
        return True

    def read_gatt_char(self, gatt_char) -> bool:
        if self._gatt_client is None:
            self.logger.error("BleInterface: No connection to read characteristic!")
            return False
        try:
            future = asyncio.run_coroutine_threadsafe(
                self._read_gatt_char(gatt_char), self._event_loop)
            success = future.result()
            if not success:
                self.logger.error("BleInterface: Failed to read characteristic!")
                return False
        except:
            self.logger.exception("BleInterface: Error when scheduling read char operation!")
            return False
        return True

    def get_gatt_profile(self) -> NaviBeltGattProfile:
        return self._gatt_profile

    # --------------------------------------------------------------- #
    # Implementation of Thread methods

    def run(self):
        try:
            # Create loop
            self._event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._event_loop)
            # Start loop
            self.logger.debug("BleInterface: Event loop started.")
            self._event_loop_ready.set()
            self._event_loop.run_forever()
        except:
            self.logger.exception("BleInterface: Error in event loop.")
            self.close()


class BleEventNotifier(threading.Thread):
    """
    A thread to notify the BLE delegate of events.
    This is necessary to isolate the BLE client that runs in an event loop.
    """

    EVENT_CONNECTION = "EVENT_CONNECTION"
    EVENT_DISCONNECTION = "EVENT_DISCONNECTION"
    EVENT_GATT_NOTIFICATION = "EVENT_GATT_NOTIFICATION"

    def __init__(self, delegate, ble_interface):
        """
        Constructor with a delegate for notifications.
        :param BeltCommunicationDelegate delegate: The delegate that handles notifications.
        :param BleInterface ble_interface: The BLE interface.
        """
        threading.Thread.__init__(self, name="BleDelegateNotifier")
        # Delegate and BLE interface
        self._delegate = delegate
        self._ble_interface = ble_interface
        # Notification queue
        self._notification_queue = queue.Queue()
        # Logger
        self.logger = logging.getLogger(__name__)

    def run(self):
        self.logger.debug("BleDelegateNotifier: Delegate notifier started.")
        while True:
            # Wait for next event
            event = self._notification_queue.get()
            if isinstance(event, tuple) and len(event) > 0:

                if event[0] == self.EVENT_CONNECTION:
                    try:
                        self._delegate.on_connection_established()
                    except:
                        self.logger.exception("BleDelegateNotifier: Unable to notify connection!")

                elif event[0] == self.EVENT_DISCONNECTION:
                    # Clear notifier reference in controller
                    try:
                        self._ble_interface._event_notifier = None
                    except:
                        self.logger.exception("BeltEventNotifier: Unable to clear event notifier reference!")
                    # Notify disconnection
                    try:
                        self._delegate.on_connection_closed(expected=event[1])
                    except:
                        self.logger.exception("BleDelegateNotifier: Unable to notify disconnection!")
                    # Stop thread
                    break

                elif event[0] == self.EVENT_GATT_NOTIFICATION:
                    try:
                        self._delegate.on_gatt_char_notified(event[1], event[2])
                    except:
                        self.logger.exception("BleDelegateNotifier: Unable to notify GATT char notification!")

                else:
                    self.logger.warning("BleDelegateNotifier: Unknown event!")
            else:
                self.logger.warning("BleDelegateNotifier: Unknown event!")

        self.logger.debug("BeltEventNotifier: Event notifier stopped.")

    def notify_gatt_notification(self, gatt_char, data):
        """
        Notifies asynchronously the delegate of a GATT notification.
        :param gatt_char: The GATT characteristic on which the notification has been received.
        :param data: The data received.
        """
        if self.is_alive():
            self._notification_queue.put((self.EVENT_GATT_NOTIFICATION, gatt_char, data))
        else:
            self.logger.warning("BeltEventNotifier: No GATT notification as the notifier thread is stopped!")

    def notify_connection(self):
        """
        Notifies asynchronously the delegate that a connection has been established.
        """
        if self.is_alive():
            self._notification_queue.put((self.EVENT_CONNECTION,))
        else:
            self.logger.warning("BeltEventNotifier: No connection notification as the notifier thread is stopped!")

    def notify_disconnection(self, expected=True, join=True):
        """
        Notifies asynchronously the delegate that the device has been disconnected, and stop the event notification
        thread.
        :param expected: True to notify that the disconnection is expected, False if it is an unexpected disconnection.
        :param join: True to wait the end of the event notification thread.
        """
        if self.is_alive():
            self._notification_queue.put((self.EVENT_DISCONNECTION, expected))
            if join:
                try:
                    self.join(THREAD_JOIN_TIMEOUT_SEC)
                except:
                    pass
        else:
            self.logger.warning("BeltEventNotifier: No disconnection notification as the notifier thread is stopped!")



# Copyright 2020, feelSpace GmbH, <info@feelspace.de>

from pybelt._communication_interface import *
from pybelt._gatt_profile import *
from typing import List

WAIT_ACK_TIMEOUT_SEC = 1  # Default timeout for waiting acknowledgment
DEBUG_MESSAGE_COMPLETION_TIMEOUT = 0.5  # Timeout for waiting the completion of a debug message


class BeltController(BeltCommunicationDelegate):
    """Belt connection and control interface.

    The belt controller supports USB and BLE connection.
    """

    # --------------------------------------------------------------- #
    # Public methods

    def __init__(self, delegate=None):
        """Initializes the belt controller.

        Parameters
        ----------
        :param BeltControllerDelegate delegate:
            The delegate that handles belt events.
        """
        # Logger
        self.logger = logging.getLogger(__name__)

        # Delegates
        self._delegate = delegate
        self._notifications_handlers = []

        # Connection state
        self._connection_state = BeltConnectionState.DISCONNECTED

        # Connection
        self._communication_interface = None
        self._last_connected_interface = None

        # Packet ack
        self._ack_char = None
        self._ack_data = None
        self._ack_event = threading.Event()

        # GATT profile
        self._gatt_profile = None

        # Cache of belt parameters
        self._belt_mode = None
        self._default_intensity = None
        self._firmware_version = None
        self._battery_status = None
        self._belt_orientation = None
        self._heading_offset = None
        self._compass_accuracy_signal_enabled = None

    def connect(self, belt):
        """ Connects a belt via Bluetooth LE or USB.

        :param Union[str, bleak.backends.device.BLEDevice] belt: The interface to use for communicating with the belt.
            For a Bluetooth LE connection a `BLEDevice` must be passed. For USB connection, the name of the serial port
            must passed, e.g. 'COM1' on Windows or '/dev/ttyUSB0' on Linux.
        :raises ValueError: When the type of interface is unsupported.
        """
        # Check belt interface
        if not isinstance(belt, (str, BLEDevice)):
            raise ValueError("Unsupported type for the belt interface.")
        # Close previous connection
        self._close_connection()
        # Set state as CONNECTING
        self._set_connection_state(BeltConnectionState.CONNECTING)
        # Open connection
        try:
            if isinstance(belt, str):
                # USB connection
                self._communication_interface = SerialPortInterface(self)
                self._communication_interface.open(belt)
            else:
                # Bluetooth connection
                self._communication_interface = BleInterface(self)
                self._communication_interface.open(belt)
        except:
            self.logger.exception("BeltController: Connection failed.")
            self._close_connection()
            self._set_connection_state(
                BeltConnectionState.DISCONNECTED,
                BeltConnectionError("Connection failed."))
            return
        # Retrieve GATT profile
        self._gatt_profile = self._communication_interface.get_gatt_profile()
        # Handshake
        if not self._handshake():
            # Handshake failed
            self._close_connection()
            self._set_connection_state(
                BeltConnectionState.DISCONNECTED,
                BeltConnectionError("Handshake failed."))
            return
        # Keep last connected interface
        self._last_connected_interface = belt
        # Set state and inform delegate
        self._set_connection_state(BeltConnectionState.CONNECTED)

    def reconnect(self):
        """
        Reconnects the last device.
        """
        if self._connection_state != BeltConnectionState.DISCONNECTED:
            self.logger.error("BeltController: Cannot reconnect when not disconnected!")
            return
        if self._last_connected_interface is None:
            self.logger.error("BeltController: No previous connection to reconnect!")
            return
        try:
            self.connect(self._last_connected_interface.get_port())
        except:
            self.logger.exception("BeltController: Error when reconnecting.")

    def disconnect_belt(self):
        """Disconnects the belt.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTING or
                self._connection_state == BeltConnectionState.DISCONNECTED):
            return
        self._set_connection_state(BeltConnectionState.DISCONNECTING)
        self._close_connection()

    def get_connection_state(self) -> int:
        """Returns the connection state.
        :return: The connection state.
        """
        return self._connection_state

    def get_belt_mode(self) -> int:
        """Returns the belt mode.
        :return: The belt mode, or None if not connected.
        """
        return self._belt_mode

    def set_belt_mode(self, mode, wait_ack=False) -> bool:
        """ Sets the mode of the belt.
        This operation is asynchronous except if the parameter ´wait_ack´ is True.
        The delegate will be informed of the mode change via ´on_belt_mode_changed´.

        :param int mode: The mode to be set. See ´BeltMode´ for a list of the modes.
        :param bool wait_ack: True to wait for mode change acknowledgment.
        :return: True if the request has been sent, False if no belt is connected.
        :raise ValueError: If the mode value is not valid.
        :raise TimeoutError: If the acknowledgment is waited and the timeout period is reached.
        """
        if self._connection_state != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: Cannot set the belt mode when not connected.")
            return False
        if mode < 0 or mode > 6:
            raise ValueError("Belt mode value out of range.")
        if wait_ack:
            write_result = self.write_gatt(
                self._gatt_profile.param_request_char,
                bytes([0x01, 0x81, mode]),
                self._gatt_profile.param_notification_char,
                b'\x01\x01')
        else:
            write_result = self.write_gatt(
                self._gatt_profile.param_request_char,
                bytes([0x01, 0x81, mode]))
        if write_result == 2:
            raise TimeoutError("Timeout period reached when changing the belt mode.")
        return write_result == 0

    def get_firmware_version(self) -> int:
        """Returns the firmware version of the connected belt.
        :return: The firmware version of the connected belt, or None if no belt is connected.
        """
        return self._firmware_version

    def get_default_intensity(self) -> int:
        """Returns the default vibration intensity of the connected belt.
        :return: The default vibration intensity of the connected belt, or None if no belt is connected.
        """
        return self._default_intensity

    def set_default_intensity(self, intensity, vibration_feedback=False, wait_ack=False) -> bool:
        """
        Sets the default intensity for belt signals such as compass.

        Note: The minimum default intensity is 5%. Any request to change the intensity below 5% will result in setting
        this minimum intensity.

        :param int intensity: The default intensity to set in range [0-100].
        :param bool vibration_feedback: True to start a vibration informing about the new intensity, False for no
        vibration feedback.
        :param bool wait_ack: True to wait for mode change acknowledgment.
        :return: 'True' if the request has been sent successfully.
        :raise TimeoutError: If the timeout period is reached when waiting for acknowledgment.
        """
        if self._connection_state != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: Cannot set the default intensity when not connected.")
            return False
        if intensity < 0:
            intensity = 0
        if intensity > 100:
            intensity = 100
        if wait_ack:
            write_result = self.write_gatt(
                self._gatt_profile.param_request_char,
                bytes([0x01, 0x82, intensity, 0x00, (0x01 if vibration_feedback else 0x00)]),
                self._gatt_profile.param_notification_char,
                b'\x01\x02')
        else:
            write_result = self.write_gatt(
                self._gatt_profile.param_request_char,
                bytes([0x01, 0x82, intensity, 0x00, (0x01 if vibration_feedback else 0x00)]))
        if write_result == 2:
            raise TimeoutError("Timeout period reached when changing the belt mode.")
        return write_result == 0

    def write_gatt(self, gatt_char, data, ack_char=None, ack_data=None, timeout_sec=WAIT_ACK_TIMEOUT_SEC) -> int:
        """
        Sends data to a GATT characteristic.

        :param GattCharacteristic gatt_char: The characteristic to write.
        :param bytes data: The data to write.
        :param GattCharacteristic ack_char: The characteristic for which an acknowledgment should be waited.
        :param bytes ack_data: The acknowledgment pattern.
        :param float timeout_sec: The timeout period in seconds.
        :return: Returns '0' if successful, '1' when no connection is available or a problem occurs, '2' when the
        timeout is reached.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            self.logger.error("BeltController: No connection to send packet.")
            return 1
        # Set ACK
        if ack_char is not None or ack_data is not None:
            self._ack_char = ack_char
            self._ack_data = ack_data
            self._ack_event.clear()
        # Send packet
        try:
            self.logger.log(5, "BeltController: " + gatt_char.uuid[4:8] + " -> " + bytes_to_hexstr(data))
        except:
            pass
        try:
            if not self._communication_interface.write_gatt_char(gatt_char, data):
                self.logger.error("BeltController: Error when sending packet.")
                self._ack_char = None
                self._ack_data = None
                self._ack_event.clear()
                return 1
        except:
            self.logger.exception("BeltController: Error when sending packet.")
            self._ack_char = None
            self._ack_data = None
            self._ack_event.clear()
            return 1
        # Wait ack
        if ack_char is not None or ack_data is not None:
            if not self._ack_event.is_set():
                # Wait for ACK
                self._ack_event.wait(timeout_sec)
            if not self._ack_event.is_set():
                # Ack not received
                self.logger.error("BeltController: ACK not received.")
                self._ack_char = None
                self._ack_data = None
                self._ack_event.clear()
                return 2
            # Clear ACK flag
            self._ack_char = None
            self._ack_data = None
            self._ack_event.clear()
        return 0

    def read_gatt(self, gatt_char, timeout_sec=WAIT_ACK_TIMEOUT_SEC) -> int:
        """
        Request the value of a GATT characteristic.

        :param GattCharacteristic gatt_char: The characteristic to read.
        :param float timeout_sec: The timeout period in seconds.
        :return: Returns '0' if successful, '1' when no connection is available or a problem occurs, '2' when the
        timeout is reached.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            self.logger.error("BeltController: No connection to send packet.")
            return 1
        # Set ACK
        self._ack_char = gatt_char
        self._ack_data = None
        self._ack_event.clear()
        # Request value
        try:
            if not self._communication_interface.read_gatt_char(gatt_char):
                self.logger.error("BeltController: Error when requesting characteristic value.")
                self._ack_char = None
                self._ack_data = None
                self._ack_event.clear()
                return 1
        except:
            self.logger.exception("BeltController: Error when requesting characteristic value.")
            self._ack_char = None
            self._ack_data = None
            self._ack_event.clear()
            return 1
        # Wait ack
        if not self._ack_event.is_set():
            # Wait for ACK
            self._ack_event.wait(timeout_sec)
        if not self._ack_event.is_set():
            # Ack not received
            self.logger.error("BeltController: ACK not received.")
            self._ack_char = None
            self._ack_data = None
            self._ack_event.clear()
            return 2
        # Clear ACK flag
        self._ack_char = None
        self._ack_data = None
        self._ack_event.clear()
        return 0

    def add_notifications_handler(self, handler):
        """
        Adds a notifications handler.
        :param BeltNotificationsHandler handler: The notifications handler to add.
        """
        self._notifications_handlers.append(handler)

    def remove_notifications_handler(self, handler):
        """
        Removes a notifications handler.
        :param BeltNotificationsHandler handler: The handler to remove
        """
        self._notifications_handlers.remove(handler)

    def set_orientation_notifications(self, enabled) -> bool:
        """
        Sets the state of orientation notifications.

        IMPORTANT: It is important to note that orientation updates from the belt were not designed for updating a
        system in real-time, just for monitoring and showing on a map the orientation of the belt. In case you need a
        vibration signal on the belt that is dependent on the orientation, you can use the `vibrate_at_magnetic_bearing`
        or `send_pulse_command` (with orientation_type=MAGNETIC_BEARING) to get a vibration relative to magnetic North.

        :param enabled: 'True' to enable orientation notifications, 'False' to disable.
        :return: 'True' if the request has been sent successfully.
        """
        if self._connection_state == BeltConnectionState.DISCONNECTED:
            return False
        return self._communication_interface.set_gatt_notifications(self._gatt_profile.orientation_data_char, enabled)

    def set_power_status_notifications(self, enabled) -> bool:
        """
        Sets the state of power status notifications.

        :param enabled: 'True' to enable power status notifications, 'False' to disable.
        :return: 'True' if the request has been sent successfully.
        """
        if self._connection_state == BeltConnectionState.DISCONNECTED:
            return False
        return self._communication_interface.set_gatt_notifications(self._gatt_profile.battery_status_char, enabled)

    def rename(self, suffix) -> bool:
        """
        Renames the belt with a suffix.
        :param str suffix: The suffix of the belt to set.
        :return: 'True' if the request has been sent successfully.
        """
        # TODO To be moved in diagnostic app
        # TODO Adds parameter wait_ack
        # TODO Raise ValueError If a parameter value is illegal.
        # TODO Reset name if suffix is None
        if self._connection_state != BeltConnectionState.CONNECTED:
            return False
        encoded_suffix = None
        try:
            encoded_suffix = suffix.encode()
        except:
            self.logger.exception("BeltController: Unable to encode the belt suffix.")
            return False
        if len(encoded_suffix) > 18:
            # Do not fit the 20 bytes max packet size
            self.logger.error("BeltController: Suffix is too long.")
            return False
        # Sent rename request
        if self.write_gatt(
                self._gatt_profile.param_request_char,
                bytes([0x01, 0x84]) + encoded_suffix) != 0:
            return False
        self.logger.debug("BeltController: Rename request sent.")
        if isinstance(self._communication_interface, BleInterface):
            self._communication_interface.close()

    def set_pairing_requirement(self, pairing_required, save=True, wait_ack=False) -> bool:
        """
        Sets the pairing requirement of the belt. The pairing requirement determines if the belt requires pairing or not
        for connections. When pairing is disabled, any device can connect to the belt without pairing. When pairing is
        enable, the belt only accept connection from paired devices. To pair a device when pairing is enabled, the
        pairing mode must be started on the belt by pressing the home button for at least 3 seconds, then the device to
        connect must request pairing.

        It is recommended to enable pairing for a general usage of the belt and to disable it only for testing purpose.

        :param bool pairing_required: 'True' to require pairing.
        :param bool save: 'True' to save the setting on the belt, False for a temporary setting.
        :param bool wait_ack: 'True' to wait for acknowledgment.
        :return: 'True' if the request has been sent successfully.
        """
        if self._connection_state != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: Cannot set pairing requirement when not connected.")
            return False
        if wait_ack:
            write_result = self.write_gatt(
                self._gatt_profile.param_request_char,
                bytes([
                    0x11,
                    0x25,
                    (0x01 if save else 0x00),
                    (0x01 if pairing_required else 0x00)]),
                self._gatt_profile.param_notification_char,
                b'\x10\x25')
        else:
            write_result = self.write_gatt(
                self._gatt_profile.param_request_char,
                bytes([
                    0x11,
                    0x25,
                    (0x01 if save else 0x00),
                    (0x01 if pairing_required else 0x00)]),
                self._gatt_profile.param_notification_char)
        if write_result == 2:
            raise TimeoutError("Timeout period reached when setting pairing requirement.")
        return write_result == 0

    def set_orientation_notification_period(self, period, start_notification) -> bool:
        """
        Sets the period for orientation notifications.

        Changing the period of orientation notifications is only available from firmware version 52. The minimum
        notification period supported by the belt is 20ms (i.e. 50Hz).

        IMPORTANT: It is important to note that orientation updates from the belt were not designed for updating a
        system in real-time, just for monitoring and showing on a map the orientation of the belt. In case you need a
        vibration signal on the belt that is dependent on the orientation, you can use the `vibrate_at_magnetic_bearing`
        or `send_pulse_command` (with orientation_type=MAGNETIC_BEARING) to get a vibration relative to magnetic North.

        The belt orientation may be inaccurate when used indoor because of magnetic interference. To get a better
        orientation, it is recommended to calibrate the belt in the room where it is used (or outdoor if used outdoor).

        It is not recommended to use short notification period over BLE connection. For short notification
        periods, after disconnection, it may be required to wait a few seconds to reconnect the belt again. If the
        reconnection failed, it may be necessary to restart the belt. In case the belt does not respond anymore, you can
        make a “hard power-off” by pressing the power button of the belt more than 6 seconds and then pressing it again
        to restart the belt.

        :param period: The period in seconds. Minimum value is 20 milliseconds (50Hz).
        :param start_notification: 'True' to start the orientation notification.
        :return: 'True' on success, 'False' otherwise.
        """
        if self.get_connection_state() != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: No belt belt connected to set notification period.")
            return False
        if self.get_firmware_version() < 52:
            self.logger.warning("BeltController: Belt firmware version not supported.")
            return False
        if period < 0.02:
            self.logger.warning("BeltController: Notification period not supported.")
            return False
        period_ms = int(period * 1000.0)
        self.set_orientation_notifications(False)
        if self.write_gatt(self._gatt_profile.param_request_char,
                           bytes([
                               0x11,
                               0x0F,
                               0x00,
                               period_ms & 0xFF,
                               (period_ms >> 8) & 0xFF,
                           ]),
                           self._gatt_profile.param_notification_char,
                           b'\x10\x0F') != 0:
            self.logger.warning("BeltController: Failed to write notification period parameter.")
            return False
        return self.set_orientation_notifications(True)

    def vibrate_at_magnetic_bearing(
            self,
            bearing,
            switch_to_app_mode=True,
            channel_index=1,
            intensity=None,
            clear_other_channels=False) -> bool:
        """
        Sends a command to start a continuous vibration in a given direction relative to magnetic North.

        :param int bearing: The direction relative to magnetic North in degrees (positive clockwise).
        :param bool switch_to_app_mode: `True` to switch automatically to app mode.
        :param int channel_index: The channel index to configure. The belt has six channels (index 0 to 5).
        :param int intensity: The intensity of the vibration in range [0, 100] or `None` to use the default intensity.
        :param bool clear_other_channels: `True` to stop and clear other channels when this vibration starts.
        :return: `True` if the command has been sent successfully.
        :raise ValueError: If a parameter value is illegal.
        """
        if self._connection_state != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: Cannot send a command when not connected.")
            return False
        if self._belt_mode != BeltMode.APP_MODE and switch_to_app_mode:
            if not self.set_belt_mode(BeltMode.APP_MODE):
                return False
        return self.send_vibration_command(
            channel_index=channel_index,
            pattern=BeltVibrationPattern.CONTINUOUS,
            intensity=intensity,
            orientation_type=BeltOrientationType.MAGNETIC_BEARING,
            orientation=bearing,
            pattern_iterations=None,
            pattern_period=500,
            pattern_start_time=0,
            exclusive_channel=False,
            clear_other_channels=clear_other_channels
        )

    def vibrate_at_angle(
            self,
            angle,
            switch_to_app_mode=True,
            channel_index=1,
            intensity=None,
            clear_other_channels=False) -> bool:
        """
        Sends a command to start a continuous vibration in a given direction relative to user front direction.

        :param int angle: The orientation of the vibration in degrees (positive clockwise), 0 is for front.
        :param bool switch_to_app_mode: `True` to switch automatically to app mode.
        :param int channel_index: The channel index to configure. The belt has six channels (index 0 to 5).
        :param int intensity: The intensity of the vibration in range [0, 100] or `None` to use the default intensity.
        :param bool clear_other_channels: `True` to stop and clear other channels when this vibration starts.
        :return: `True` if the command has been sent successfully.
        :raise ValueError: If a parameter value is illegal.
        """
        if self._connection_state != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: Cannot send a command when not connected.")
            return False
        if self._belt_mode != BeltMode.APP_MODE and switch_to_app_mode:
            if not self.set_belt_mode(BeltMode.APP_MODE):
                return False
        return self.send_vibration_command(
            channel_index=channel_index,
            pattern=BeltVibrationPattern.CONTINUOUS,
            intensity=intensity,
            orientation_type=BeltOrientationType.ANGLE,
            orientation=angle,
            pattern_iterations=None,
            pattern_period=500,
            pattern_start_time=0,
            exclusive_channel=False,
            clear_other_channels=clear_other_channels
        )

    def send_vibration_command(
            self,
            channel_index,
            pattern,
            intensity,
            orientation_type,
            orientation,
            pattern_iterations,
            pattern_period,
            pattern_start_time,
            exclusive_channel,
            clear_other_channels) -> bool:
        """
        Sends a command that configures the vibration on a vibration channel.

        :param int channel_index: The channel index to configure. The belt has six channels (index 0 to 5).
        :param int pattern: The vibration pattern to use, see `BeltVibrationPattern`.
        :param Union[int,None] intensity: The intensity of the vibration in range [0, 100] or `None` to use the default
            intensity.
        :param int orientation_type: The type of signal orientation, see `BeltOrientationType`.
        :param int orientation: The value of the vibration orientation.
        :param Union[int,None] pattern_iterations: The number of pattern iterations or `None` to repeat indefinitely the
            pattern. The maximum value is 127 iterations.
        :param int pattern_period: The duration in milliseconds of one pattern iteration. The maximum period is 65535
            milliseconds.
        :param int pattern_start_time: The starting time in milliseconds of the first pattern iteration.
        :param bool exclusive_channel: `True` to suspend other channels as long as this vibration is active.
        :param bool clear_other_channels: `True` to stop and clear other channels when this vibration starts.
        :return: `True` if the command has been sent successfully.
        :raise ValueError: If a parameter value is illegal.
        """
        if channel_index < 0 or channel_index > 5:
            raise ValueError("Channel index value out of range.")
        if pattern < 0 or pattern > 26:
            raise ValueError("Pattern value out of range.")
        if (intensity is not None) and (intensity < 0 or intensity > 100):
            raise ValueError("Intensity value out of range.")
        if orientation_type < 0 or orientation_type > 3:
            raise ValueError("Orientation type value out of range.")
        if (pattern_iterations is not None) and (pattern_iterations < 0 or pattern_iterations > 127):
            raise ValueError("Pattern iterations value out of range.")
        if pattern_period <= 0 or pattern_period > 65535:
            raise ValueError("Pattern period value out of range.")
        if pattern_start_time < 0 or pattern_start_time > 65535:
            raise ValueError("Pattern start time value out of range.")
        if self._connection_state != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: Cannot send a command when not connected.")
            return False
        # Adjust values
        if intensity is None:
            intensity = 0xAAAA
        if orientation_type == BeltOrientationType.MAGNETIC_BEARING or orientation_type == BeltOrientationType.ANGLE:
            orientation = orientation % 360
        if orientation_type == BeltOrientationType.MOTOR_INDEX:
            orientation = orientation % 16
        # Send command
        return self.write_gatt(
            self._gatt_profile.vibration_command_char,
            bytes([
                channel_index,
                pattern,
                intensity & 0xFF,
                (intensity >> 8) & 0xFF,
                0x00,
                0x00,
                orientation_type,
                orientation & 0xFF,
                (orientation >> 8) & 0xFF,
                0x00,
                0x00,
                (0x00 if pattern_iterations is None else pattern_iterations),
                pattern_period & 0xFF,
                (pattern_period >> 8) & 0xFF,
                pattern_start_time & 0xFF,
                (pattern_start_time >> 8) & 0xFF,
                (0x01 if exclusive_channel else 0x00),
                (0x01 if clear_other_channels else 0x00)
            ])) == 0

    def send_pulse_command(
            self,
            channel_index,
            orientation_type,
            orientation,
            intensity,
            on_duration_ms,
            pulse_period,
            pulse_iterations,
            series_period,
            series_iterations,
            timer_option,
            exclusive_channel,
            clear_other_channels) -> bool:
        """
        Sends a command that configures vibration pulses on a vibration channel.

        :param int channel_index: The channel index to configure. The belt has six channels (index 0 to 5).
        :param int orientation_type: The type of signal orientation, see `BeltOrientationType`.
        :param int orientation: The value of the vibration orientation.
        :param Union[int,None] intensity: The intensity of the vibration in range [0, 100] or `None` to use the default
            intensity.
        :param int on_duration_ms: The on-duration of a pulse in milliseconds.
        :param int pulse_period: The period of pulses in milliseconds.
        :param int pulse_iterations: The number of pulses in a series.
        :param int series_period: The period of a series of pulses.
        :param Union[int,None] series_iterations: The number of series iterations or `None` to repeat the series of
            pulse indefinitely.
        :param int timer_option: Behavior of the timer for vibration, see `BeltVibrationTimerOption`.
        :param bool exclusive_channel: `True` to suspend other channels as long as this vibration is active.
        :param bool clear_other_channels: `True` to stop and clear other channels when this vibration starts.
        :return: `True` if the command has been sent successfully.
        :raise ValueError: If a parameter value is illegal.
        """
        if channel_index < 0 or channel_index > 5:
            raise ValueError("Channel index value out of range.")
        if orientation_type < 0 or orientation_type > 3:
            raise ValueError("Orientation type value out of range.")
        if (intensity is not None) and (intensity < 0 or intensity > 100):
            raise ValueError("Intensity value out of range.")
        if on_duration_ms <= 0 or on_duration_ms > 65535:
            raise ValueError("On-duration value out of range.")
        if pulse_period <= 0 or pulse_period > 65535:
            raise ValueError("On-duration value out of range.")
        if pulse_iterations <= 0 or pulse_iterations > 255:
            raise ValueError("Pulse iterations value out of range.")
        if series_period <= 0 or series_period > 65535:
            raise ValueError("Series period value out of range.")
        if (series_iterations is not None) and (series_iterations < 0 or series_iterations > 127):
            raise ValueError("Series iterations value out of range.")
        if timer_option < 0 or timer_option > 2:
            raise ValueError("Timer option value out of range.")
        # Adjust values
        if intensity is None:
            intensity = 0xAA
        if orientation_type == BeltOrientationType.MAGNETIC_BEARING or orientation_type == BeltOrientationType.ANGLE:
            orientation = orientation % 360
        if orientation_type == BeltOrientationType.MOTOR_INDEX:
            orientation = orientation % 16
        # Send command
        return self.write_gatt(
            self._gatt_profile.vibration_command_char,
            bytes([
                0x40,
                channel_index,
                orientation_type,
                orientation & 0xFF,
                (orientation >> 8) & 0xFF,
                intensity,
                on_duration_ms & 0xFF,
                (on_duration_ms >> 8) & 0xFF,
                pulse_iterations,
                (0x00 if series_iterations is None else series_iterations),
                pulse_period & 0xFF,
                (pulse_period >> 8) & 0xFF,
                series_period & 0xFF,
                (series_period >> 8) & 0xFF,
                timer_option,
                (0x01 if exclusive_channel else 0x00),
                (0x01 if clear_other_channels else 0x00)
            ])) == 0

    def stop_vibration(
            self,
            channel_index=None) -> bool:
        """
        Stops the vibration on all or one vibration channel in App mode.

        :param int channel_index: The channel index to stop in range [0, 5], or `None` to stop all channels.
        :return: `True` if the command has been sent successfully.
        :raise ValueError: If the channel index value is out of range.
        """
        if (channel_index is not None) and (channel_index < 0 or channel_index > 5):
            raise ValueError("Channel index value out of range.")
        if self._connection_state != BeltConnectionState.CONNECTED:
            self.logger.warning("BeltController: Cannot send a command when not connected.")
            return False
        if channel_index is None:
            return self.write_gatt(
                self._gatt_profile.vibration_command_char,
                bytes([0x30, 0xFF])) == 0
        else:
            return self.write_gatt(
                self._gatt_profile.vibration_command_char,
                bytes([0x30, channel_index & 0xFF])) == 0

    # --------------------------------------------------------------- #
    # Private methods

    def _set_connection_state(self, state, error=None, notify=True):
        """Sets the connection state.
        :param int state: The state to be set.
        :param Exception error: The error to notify if any.
        :param bool notify: `True` to notify the delegate.
        """
        if self._connection_state == state:
            return
        self._connection_state = state
        if notify:
            try:
                self._delegate.on_connection_state_changed(state, error=error)
            except:
                pass

    def _close_connection(self):
        """Closes the connection and clear cached parameter values.
        The connection state is not changed and delegate is not informed.
        """
        if self._communication_interface is not None:
            self._communication_interface.close()
        self._belt_mode = None
        self._default_intensity = None
        self._firmware_version = None
        self._battery_status = None
        self._belt_orientation = None
        self._heading_offset = None
        self._compass_accuracy_signal_enabled = None

    def _handshake(self):
        """Handshake procedure.

        :return: 'True' if the handshake is successful, 'False' otherwise.
        """
        self.logger.info("BeltController: Start handshake.")
        # Register to keep-alive
        self.logger.debug("BeltController: Register to keep-alive notifications.")
        if not self._communication_interface.set_gatt_notifications(self._gatt_profile.keep_alive_char, True):
            return False

        # Register to parameter notifications
        self.logger.debug("BeltController: Register to parameter notifications.")
        if not self._communication_interface.set_gatt_notifications(self._gatt_profile.param_notification_char, True):
            return False

        # Read belt mode
        self.logger.debug("BeltController: Read belt mode.")
        if (self.write_gatt(self._gatt_profile.param_request_char,
                            b'\x01\x01',
                            self._gatt_profile.param_notification_char,
                            b'\x01\x01') != 0):
            self.logger.error("BeltController: Failed to request belt mode.")
            return False
        if self._belt_mode is None:
            self.logger.error("BeltController: Failed to read belt mode.")
            return False

        # Read default intensity
        self.logger.debug("BeltController: Read default intensity.")
        if (self.write_gatt(self._gatt_profile.param_request_char,
                            b'\x01\x02',
                            self._gatt_profile.param_notification_char,
                            b'\x01\x02') != 0):
            self.logger.error("BeltController: Failed to request default intensity.")
            return False
        if self._default_intensity is None:
            self.logger.error("BeltController: Failed to read default intensity.")
            return False

        # Read firmware version
        self.logger.debug("BeltController: Read firmware version.")
        if self.read_gatt(self._gatt_profile.firmware_info_char) != 0:
            self.logger.error("BeltController: Failed to request firmware version.")
            return False
        if self._firmware_version is None:
            self.logger.error("BeltController: Failed to read firmware version.")
            return False

        # Read heading offset
        self.logger.debug("BeltController: Read heading offset.")
        if (self.write_gatt(self._gatt_profile.param_request_char,
                            b'\x01\x03',
                            self._gatt_profile.param_notification_char,
                            b'\x01\x03') != 0):
            self.logger.error("BeltController: Failed to request default intensity.")
            return False
        if self._heading_offset is None:
            self.logger.error("BeltController: Failed to read default intensity.")
            return False

        # Read compass accuracy signal state
        self.logger.debug("BeltController: Read compass accuracy signal state.")
        if (self.write_gatt(self._gatt_profile.param_request_char,
                            b'\x10\x01\x03',
                            self._gatt_profile.param_notification_char,
                            b'\x10\x03') != 0):
            self.logger.error("BeltController: Failed to request compass accuracy signal state.")
            return False
        if self._compass_accuracy_signal_enabled is None:
            self.logger.error("BeltController: Failed to read compass accuracy signal state.")
            return False

        # Register to button press
        self.logger.debug("BeltController: Register to button press events.")
        if not self._communication_interface.set_gatt_notifications(self._gatt_profile.button_press_char, True):
            self.logger.error("BeltController: Failed to register to button press events.")
            return False

        # Register to orientation notifications
        self.logger.debug("BeltController: Register to orientation notifications.")
        if not self._communication_interface.set_gatt_notifications(self._gatt_profile.orientation_data_char, True):
            self.logger.error("BeltController: Failed to register to orientation notifications.")
            return False

        # Register to power status notifications
        self.logger.debug("BeltController: Register to power-status notifications.")
        if not self._communication_interface.set_gatt_notifications(self._gatt_profile.battery_status_char, True):
            self.logger.error("BeltController: Failed to register to power-status notifications.")
            return False

        self.logger.info("BeltController: Handshake completed.")
        return True

    def _set_belt_mode(self, belt_mode):
        """Sets the belt mode and informs the delegate.

        :param int belt_mode:
            The belt mode to set.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        if belt_mode < 0 or belt_mode > 8:
            self.logger.error("BeltController: Illegal belt mode.")
            return
        if belt_mode == self._belt_mode:
            return
        self._belt_mode = belt_mode
        if self._connection_state == BeltConnectionState.CONNECTING:
            # No delegate notification during handshake
            return
        try:
            self._delegate.on_belt_mode_changed(belt_mode)
        except:
            pass

    def _notify_button_pressed(self, button_id, previous_mode, new_mode):
        """
        Notifies the delegate of a button press event and sets the belt mode.

        :param int button_id:
            The button ID.
        :param int previous_mode:
            The belt mode before the press event.
        :param int new_mode:
            The belt mode after the press event.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        if previous_mode < 0 or previous_mode > 6 or new_mode < 0 or new_mode > 6 or button_id < 1 or button_id > 4:
            self.logger.error("BeltController: Illegal button press event argument.")
            return
        self._belt_mode = new_mode
        try:
            self._delegate.on_belt_button_pressed(button_id, previous_mode, new_mode)
        except:
            pass

    def _notify_default_intensity(self, intensity):
        """
        Sets the default intensity member variable and notifies the delegate.

        :param intensity: The default intensity to set.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        if intensity < 0 or intensity > 100:
            self.logger.error("BeltController: Illegal intensity notification argument.")
            return
        self._default_intensity = intensity
        try:
            self._delegate.on_default_intensity_changed(intensity)
        except:
            pass

    def _notify_heading_offset(self, offset):
        """
        Sets the heading offset member variable and notifies the delegate.

        :param int offset: The offset value.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        if offset < 0 or offset > 359:
            self.logger.error("BeltController: Illegal offset notification argument.")
            return
        self._heading_offset = offset
        try:
            self._delegate.on_heading_offset_notified(offset)
        except:
            pass

    def _notify_bt_name(self, bt_name):
        """
        Notifies the delegate of the BT name.

        :param bytearray bt_name: The BT name.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        bt_name = decode_ascii(bt_name)
        try:
            self._delegate.on_bt_name_notified(bt_name)
        except:
            pass

    def _notify_compass_accuracy_signal_state(self, state):
        """
        Sets the compass accuracy signal state member variable and notifies the delegate.

        :param int state: The signal state.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        self._compass_accuracy_signal_enabled = state != 0
        try:
            self._delegate.on_compass_accuracy_signal_state_notified(state != 0)
        except:
            pass

    def _notify_pairing_requirement(self, pairing_required):
        """
        Notifies the delegate of the pairing requirement state.
        :param pairing_required: 'True' if pairing is required.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        try:
            self._delegate.on_pairing_requirement_notified(pairing_required)
        except:
            pass

    def _notify_belt_orientation(self, packet):
        """Notifies the belt orientation to the delegate.

        :param bytes packet: The raw orientation data.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        sensor_id = int.from_bytes(
            bytes(packet[0:1]),
            byteorder='little',
            signed=False)
        belt_heading = int.from_bytes(
            bytes(packet[1:3]),
            byteorder='little',
            signed=True)
        box_heading = int.from_bytes(
            bytes(packet[3:5]),
            byteorder='little',
            signed=True)
        box_roll = int.from_bytes(
            bytes(packet[5:7]),
            byteorder='little',
            signed=True)
        box_pitch = int.from_bytes(
            bytes(packet[7:9]),
            byteorder='little',
            signed=True)
        accuracy = int.from_bytes(
            bytes(packet[9:11]),
            byteorder='little',
            signed=False)
        mag_stat = int.from_bytes(
            bytes(packet[11:12]),
            byteorder='little',
            signed=True)
        acc_stat = int.from_bytes(
            bytes(packet[12:13]),
            byteorder='little',
            signed=True)
        gyr_stat = int.from_bytes(
            bytes(packet[13:14]),
            byteorder='little',
            signed=True)
        fus_stat = int.from_bytes(
            bytes(packet[14:15]),
            byteorder='little',
            signed=True)
        is_orientation_accurate = (int.from_bytes(
            bytes(packet[15:16]),
            byteorder='little',
            signed=False)) == 0
        try:
            self._delegate.on_belt_orientation_notified(
                belt_heading,
                is_orientation_accurate,
                [sensor_id,
                 belt_heading,
                 box_heading,
                 box_roll,
                 box_pitch,
                 accuracy,
                 mag_stat,
                 acc_stat,
                 gyr_stat,
                 fus_stat,
                 is_orientation_accurate])
        except:
            pass

    def _notify_belt_battery(self, packet):
        """Notifies the belt battery status to the delegate.

        :param bytes packet: The raw battery status data.
        """
        if (self._connection_state == BeltConnectionState.DISCONNECTED or
                self._connection_state == BeltConnectionState.DISCONNECTING):
            return
        bat_stat = int.from_bytes(
            bytes(packet[0:1]),
            byteorder='little',
            signed=False)
        charge_level = float(int.from_bytes(
            bytes(packet[1:3]),
            byteorder='little',
            signed=False)) / 256.0
        if charge_level > 100.0:
            charge_level = 100.0
        ttfe = float(int.from_bytes(
            bytes(packet[3:5]),
            byteorder='little',
            signed=False)) * 5.625
        ma = int.from_bytes(
            bytes(packet[5:7]),
            byteorder='little',
            signed=True)
        mv = int.from_bytes(
            bytes(packet[7:9]),
            byteorder='little',
            signed=False)
        try:
            self._delegate.on_belt_battery_notified(
                charge_level,
                [bat_stat, charge_level, ttfe, ma, mv])
        except:
            pass

    def _is_ack(self, gatt_char, data) -> bool:
        """
        Checks if the data corresponds to the current acknowledgment.

        :param GattCharacteristic gatt_char: The GATT characteristic of the notification.
        :param bytes data: The data received.
        :return: 'True' if the ACK is verified, 'False' otherwise.
        """
        try:
            if self._ack_char is not None:
                if self._ack_char != gatt_char:
                    return False
            if self._ack_data is not None:
                if len(self._ack_data) > len(data):
                    return False
                for packet_byte, ack_byte in zip(data, self._ack_data):
                    if ack_byte is not None:
                        if ack_byte != packet_byte:
                            return False
            return True
        except:
            self.logger.warning("BeltController: Unable to check ACK.")
        return False

    # --------------------------------------------------------------- #
    # Implementation of communication interface delegate methods

    def on_connection_established(self):
        # Save last connection for reconnection
        self._last_connected_interface = self._communication_interface
        self.logger.debug("BeltController: Connection established.")
        pass

    def on_connection_closed(self, expected=True):
        self.logger.debug("BeltController: Connection closed.")
        if expected or self._connection_state == BeltConnectionState.DISCONNECTING:
            self._set_connection_state(BeltConnectionState.DISCONNECTED)
        else:
            self._set_connection_state(
                BeltConnectionState.DISCONNECTED,
                BeltConnectionError("Connection lost."))

    def on_gatt_char_notified(self, gatt_char, data):

        # TODO To be moved in diagnosis app using system handler
        # Process packet
        # try:
        #     self.logger.log(5, "BeltController: "+gatt_char.uuid[4:8]+" <- "+bytes_to_hexstr(data))
        # except:
        #     pass

        # Check for service discovery completed
        if self._gatt_profile is None:
            self.logger.debug("BeltController: Notification received before service discovery.")
            return

        # Check for power-off notification
        if ((gatt_char == self._gatt_profile.button_press_char and len(data) >= 5 and data[4] == BeltMode.STANDBY) or
                (gatt_char == self._gatt_profile.param_notification_char and len(data) >= 3 and data[0] == 0x01 and
                 data[1] == 0x01 and data[2] == BeltMode.STANDBY)):
            self.logger.info("BeltController: Belt switched off.")
            self._communication_interface.close()

        # Firmware information
        if gatt_char == self._gatt_profile.firmware_info_char:
            # Firmware information received
            if len(data) >= 2:
                try:
                    self._firmware_version = int.from_bytes(
                        bytes(data[:2]), byteorder='little', signed=False)
                except:
                    self.logger.error("Unable to parse firmware version.")

        # Keep alive request
        if gatt_char == self._gatt_profile.keep_alive_char:
            # Retrieve belt mode
            if len(data) >= 2:
                self._set_belt_mode(data[1])
            # Send keep-alive ACK
            self.write_gatt(self._gatt_profile.keep_alive_char, bytes([0x01]))

        # Button press notification
        if gatt_char == self._gatt_profile.button_press_char:
            if len(data) >= 5:
                self._notify_button_pressed(data[0], data[3], data[4])

        # Belt mode change
        if gatt_char == self._gatt_profile.param_notification_char:
            if len(data) >= 3 and data[0] == 0x01 and data[1] == 0x01:
                self._set_belt_mode(data[2])

        # Default intensity
        if gatt_char == self._gatt_profile.param_notification_char:
            if len(data) >= 3 and data[0] == 0x01 and data[1] == 0x02:
                self._notify_default_intensity(data[2])

        # Heading offset
        if gatt_char == self._gatt_profile.param_notification_char:
            if len(data) >= 4 and data[0] == 0x01 and data[1] == 0x03:
                self._notify_heading_offset(int.from_bytes(
                    bytes(data[2:4]), byteorder='little', signed=False))

        # BT name
        if gatt_char == self._gatt_profile.param_notification_char:
            if len(data) >= 2 and data[0] == 0x01 and data[1] == 0x04:
                self._notify_bt_name(bytearray(data[2:]))

        # Advanced parameters
        if gatt_char == self._gatt_profile.param_notification_char:
            if len(data) >= 2 and data[0] == 0x10:
                if data[1] == 0x00:
                    # Default intensity
                    self._notify_default_intensity(data[2])
                elif data[1] == 0x01:
                    # Heading offset
                    self._notify_heading_offset(int.from_bytes(
                        bytes(data[2:4]), byteorder='little', signed=False))
                elif data[1] == 0x03:
                    # Compass accuracy signal state
                    self._notify_compass_accuracy_signal_state(data[2])
                elif data[1] == 0x25:
                    # Pairing requirement
                    self._notify_pairing_requirement(data[2] != 0)

        # Belt orientation
        if gatt_char == self._gatt_profile.orientation_data_char:
            if len(data) >= 16:
                self._notify_belt_orientation(data)

        # Battery status
        if gatt_char == self._gatt_profile.battery_status_char:
            if len(data) >= 9:
                self._notify_belt_battery(data)

        # TODO To be moved in diagnosis app using system handler
        # # Error notification
        # if gatt_char == navibelt_debug_output_char:
        #     if len(data) >= 5 and data[0] == 0xA0:
        #         error_id = int.from_bytes(bytes(data[1:5]), byteorder='little', signed=False)
        #         self.logger.error("BeltController: Belt error " + hex(error_id) + " !")

        # TODO To be moved in diagnosis app using system handler
        # # Debug message
        # if len(self._debug_message_buffer) > 0 and \
        #         time.perf_counter()-self._debug_message_last_received > DEBUG_MESSAGE_COMPLETION_TIMEOUT:
        #     # Output incomplete debug message
        #     self.logger.debug("Belt debug message: " + self._debug_message_buffer)
        #     self._debug_message_buffer = ""
        # if gatt_char == navibelt_debug_output_char:
        #     if len(data) > 1 and data[0] == 0x01:
        #         self._debug_message_buffer += decode_ascii(data[1:])
        #         self._debug_message_last_received = time.perf_counter()
        # # Output message ending with '\n'
        # eol = self._debug_message_buffer.find('\n')
        # while eol >= 0:
        #     line = self._debug_message_buffer[:eol]
        #     self._debug_message_buffer = self._debug_message_buffer[eol+1:]
        #     self.logger.debug("Belt debug message: " + line)
        #     eol = self._debug_message_buffer.find('\n')

        # TODO Other notifications

        # Check for ACK
        if (self._ack_data is not None or self._ack_char is not None) and not self._ack_event.is_set():
            if self._is_ack(gatt_char, data):
                self._ack_data = None
                self._ack_char = None
                self._ack_event.set()

        # Inform system handlers
        for handler in self._notifications_handlers:
            try:
                handler.on_gatt_char_notified(gatt_char, data)
            except:
                pass


def bytes_to_hexstr(data) -> str:
    """
    Returns a string representation of bytes.
    :param bytes data: The data to convert.
    :return: A string representation of the data.
    """
    if len(data) == 0:
        return "[]"
    strRep = "0x"
    for d in data:
        if d < 16:
            strRep = strRep + "0" + hex(d)[2:]
        else:
            strRep = strRep + hex(d)[2:]
    return strRep


class BeltConnectionState:
    """Enumeration of connection state."""

    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2
    DISCONNECTING = 3


class BeltMode:
    """Enumeration of belt modes."""

    STANDBY = 0
    WAIT = 1
    COMPASS = 2
    APP_MODE = 3
    PAUSE = 4
    CALIBRATION = 5
    CROSSING = 6


class BeltOrientationType:
    """Enumeration of orientation types."""

    BINARY_MASK = 0
    MOTOR_INDEX = 1
    ANGLE = 2
    MAGNETIC_BEARING = 3


class BeltVibrationPattern:
    """Enumeration of vibration patterns."""

    NO_VIBRATION = 0
    CONTINUOUS = 1
    SINGLE_SHORT = 2
    SINGLE_LONG = 3
    DOUBLE_SHORT = 4
    DOUBLE_LONG = 5


class BeltVibrationTimerOption:
    """Enumeration of timer option for vibration pulses commands."""

    RESET_TIMER = 0
    RESET_ON_DIFFERENT_PERIOD = 1
    KEEP_TIMER = 2


class BeltControllerDelegate:
    """Delegate interface for the belt controller.
    """

    def on_connection_state_changed(self, state, error=None):
        """Called when the connection state changed.

        Parameters
        ----------
        :param int state:
            The connection state.
        :param Exception error:
            Error if the state change is unexpected.
        """
        pass

    def on_belt_mode_changed(self, belt_mode):
        """Called when the belt mode changed.

        :param int belt_mode:
            The belt mode.
        """
        pass

    def on_belt_button_pressed(self, button_id, previous_mode, new_mode):
        """Called when a button on the belt has been pressed.

        :param int button_id:
            The button ID.
        :param int previous_mode:
            The belt mode before the button press event.
        :param int new_mode:
            The belt mode after the button press event.
        """
        pass

    def on_default_intensity_changed(self, intensity):
        """Called when the default intensity has been changed or notified.

        :param int intensity:
            The default vibration intensity of the belt.
        """
        pass

    def on_heading_offset_notified(self, offset):
        """Called when the heading offset of the belt has been changed or notified.

        :param int offset:
            The heading offset of the belt.
        """
        pass

    def on_bt_name_notified(self, bt_name):
        """Called when the BT name of the belt has been changed or notified.

        :param str bt_name:
            The BT name of the belt.
        """
        pass

    def on_compass_accuracy_signal_state_notified(self, enabled):
        """
        Called when the state of the compass accuracy signal has been changed or notified.

        :param bool enabled:
            'True' if the signal is enabled, 'False' otherwise.
        """
        pass

    def on_pairing_requirement_notified(self, pairing_required):
        """ Called when the pairing requirement has been changed or notified.

        :param pairing_required: 'True' if the pairing is required.
        """
        pass

    def on_belt_orientation_notified(self, heading, is_orientation_accurate, extra):
        """ Called when the orientation of the belt has been notified.

        :param int heading: The heading of the belt in degrees.
        :param bool is_orientation_accurate: `True` if the orientation is accurate, `False` otherwise.
        :param List extra: A list containing extra information on the orientation notification.
        """
        pass

    def on_belt_battery_notified(self, charge_level, extra):
        """ Called when the battery level of the belt is notified.
        :param float charge_level: Charge level of the belt battery in percent.
        :param List extra: Extra information on the belt battery.
        """
        pass


class BeltConnectionError(Exception):
    """Exception raised for belt connection problems.
    """

    def __init__(self, message):
        """Constructor.

        Parameters
        ----------
        :param str message:
            The error message.
        """
        super().__init__(message)


class BeltNotificationsHandler:
    """
    Handler interface for belt debug and test.
    """

    def on_gatt_char_notified(self, gatt_char, data):
        """
        Called when a GATT notification has been received or a characteristic has been read.

        :param GattCharacteristic gatt_char: The GATT characteristic.
        :param bytes data: The data received.
        """
        pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pybelt",
    version="1.1.0",
    author="feelSpace GmbH",
    author_email="dev@feelspace.de",
    description="An Python library to control the feelSpace naviBelt from your application",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/feelSpace/pybelt",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.3',
    install_requires=["pyserial", "bleak"],
)


# Copyright 2020, feelSpace GmbH, <info@feelspace.de>

import asyncio
import logging
import threading
from typing import List
from bleak.backends.device import BLEDevice
from bleak import BleakScanner
from contextlib import contextmanager


@contextmanager
def create():
    scanner = None
    try:
        scanner = BeltScanner()
        yield scanner
    finally:
        if scanner is not None:
            scanner.close()


class BeltScanner:
    """Utility class for scanning belts.
    """

    def __init__(self, event_loop=None):
        """Initializes the belt scanner.
        :param event_loop: Optional AsyncIO event loop.
        """
        self._logger = logging.getLogger(__name__)
        if event_loop is None:
            # Start scanner own event loop
            self._event_loop = asyncio.new_event_loop()
            self._event_loop_thread = _EventLoopThread(self._event_loop)
            self._event_loop_thread.start()
        else:
            self._event_loop = event_loop
            self._event_loop_thread = None

    def __del__(self):
        if self._event_loop_thread is not None:
            # Stop event loop if scanner own it
            self._event_loop.call_soon_threadsafe(self._event_loop.stop)
            self._event_loop_thread.join()
            self._event_loop.close()

    def close(self):
        if self._event_loop_thread is not None:
            # Stop event loop if scanner own it
            try:
                self._event_loop.call_soon_threadsafe(self._event_loop.stop)
                self._event_loop_thread.join()
                self._event_loop.close()
                self._event_loop_thread = None
            except:
                pass

    def scan(self) -> List[BLEDevice]:
        """Scans for advertising belts.

        :return: The available belts.
        """
        future = asyncio.run_coroutine_threadsafe(self._scan(), self._event_loop)
        return future.result()

    async def _scan(self) -> List[BLEDevice]:
        """Scans for advertising belts (asynchronous).
        """
        self._logger.debug("BeltScanner: Start async scan.")
        belts = []
        devices = await BleakScanner.discover(return_adv=True)
        for d in devices.values():
            self._logger.debug("BeltScanner: Device found.")
            # Check for service UUID
            adv_data = d[1]
            for uuid in adv_data.service_uuids:
                self._logger.debug("BeltScanner: Advertised UUID {}.".format(uuid))
                if isinstance(uuid, str) and ("65333333-a115-11e2-9e9a-0800200ca100" in uuid.lower()
                                              or "0000fe51-0000-1000-8000-00805f9b34fb" in uuid.lower()):
                    # Check if belt not yet discovered
                    belt_already_discovered = False
                    for b in belts:
                        if isinstance(b.address, str) and isinstance(d[0].address, str) and \
                                (b.address.lower() == d[0].address.lower()):
                            belt_already_discovered = True
                            break
                    if not belt_already_discovered:
                        belts.append(d[0])
        self._logger.debug("BeltScanner: End async scan.")
        return belts


class _EventLoopThread(threading.Thread):
    """Thread for the event loop.
    """

    def __init__(self, event_loop):
        """Initializes the thread for the event loop.
        :param event_loop: The event loop.
        """
        self._logger = logging.getLogger(__name__)
        threading.Thread.__init__(self, name="_EventLoopThread")
        self._event_loop = event_loop

    def run(self) -> None:
        self._logger.debug("BeltScanner: Start scan event loop.")
        self._event_loop.run_forever()
        self._logger.debug("BeltScanner: End scan event loop.")
